#
# Generate the observation count using the Spoke model
#

import numpy as np
import scipy.special

def trueInteraction(t, s, fnRate):
    return scipy.special.binom(t,s)*np.power(fnRate,t-s)*np.power(1.0 - fnRate,s)

def falseInteraction(t, s, fpRate):
    return scipy.special.binom(t,s)*np.power(1.0 - fpRate,t-s)*np.power(fpRate,s)

def interactionProbability(rho, fnRate, fpRate):
    MAX_TRIALS = 100

    mPreComputed = np.zeros(shape=(MAX_TRIALS,MAX_TRIALS), dtype=float)
    for t in np.arange(MAX_TRIALS):
        for s in np.arange(t+1):
            mPreComputed[t][s] = np.log(trueInteraction(t,s,fnRate))
            mPreComputed[t][s] -= np.log(trueInteraction(t,s,fnRate)*rho + falseInteraction(t,s,fpRate)*(1.0 - rho))

    return mPreComputed
    

class CountSpokeModel:
    
    def __init__(self, nProteins, bait_inds, incidence):

        self.nProteins = nProteins
        self.mObserved = np.zeros(shape=(nProteins, nProteins), dtype=int)
        for i, bait in zip(range(len(bait_inds)), bait_inds):
            self.mObserved[bait,:] += incidence[i,:]
            self.mObserved[:,bait] += incidence[i,:]

        self.mTrials = np.zeros(shape=(nProteins, nProteins), dtype=int)
        for bait in bait_inds:
            for j in range(nProteins):
                self.mTrials[bait][j] += 1
                self.mTrials[j][bait] += 1
        
        for i in range(nProteins):
            assert(np.sum(self.mTrials[i,:]) == np.sum(self.mTrials[:,i]))

        self.mPreComputed = interactionProbability(0.3, 0.4, 0.01)
        self.mPosterior = np.zeros(shape=(nProteins, nProteins), dtype=float)

        for i in np.arange(nProteins):
            for j in np.arange(i+1):
                t = self.mTrials[i][j]
                s = self.mObserved[i][j]
                self.mPosterior[i][j] =self.mPreComputed[t][s]

        #
        # Create the adjacency list
        #
        self.lstAdjacency = {}
        for i in np.arange(nProteins):
            self.lstAdjacency[i] = []
            for j in np.arange(nProteins):
                t = self.mTrials[i][j]
                if (i < j):
                    s = self.mObserved[i][j] 
                else:
                    s = self.mObserved[j][i] 
                if (i != j and t > 0):
                    assert(s <= t)
                    self.lstAdjacency[i].append(j)
    
    def write2cytoscape(self, indicators, matQ, vecProteins):
        nRows, nCols = matQ.shape
        with open("out.sif", "w") as fh:
            for i in np.arange(nRows):
                for j in self.lstAdjacency[i]:
                    t = self.mTrials[i][j]
                    if (i >= j):
                        continue
                    if (t > 0 and indicators[i] == indicators[j]):
                        fh.write(vecProteins[i] + '\t' + str(indicators[i]) + '\t' + str(vecProteins[j]) + '\n')
            fh.close()


#!/usr/bin/env python3

#
# BRAID: A python program to find protein complexes from high-throughput data
# 
# Input file: a CSV file containing a list of bait-preys experiments
# The first protein at the beginning of every line is a bait protein.
#   
# Model parameters:
# 1. False negative rate
# 2. False positive rate
# 3. A maximum possible number of clusters

import argparse
import numpy as np
import matplotlib.pyplot as plt
from scipy import stats

import braidmrf.observation.countSpokeModel as cpm
import braidmrf.observation.countBioplexSpokeModel as cpmBioplex
import braidmrf.meanfield.simulateLikelihood as smlt

import braidmrf.inputFile.inputFile as inputFile
import braidmrf.inputFile.inputBioplex as inputBioplex

from time import time as timer

def clustering(inputSet, Nk, psi):
    fn = 0.8
    fp = 0.04
    nProteins = inputSet.observationG.nProteins
    cmfa = smlt.CMeanFieldAnnealing(nProteins, Nk)
    ts = timer()
    cmfa.estimate(inputSet.observationG, nProteins, Nk, psi)
    te = timer()
    print("Time running MFA: ", te-ts)

    (fn, fp) = cmfa.computeResidues(inputSet.observationG, nProteins, Nk)
    cmfa.computeEntropy(nProteins, Nk)
    
    print("False negative rate = " + str(fn))
    print("False positive rate = " + str(fp))
    
    inputSet.writeCluster2File(cmfa.mIndicatorQ, cmfa.indicatorVec)
    inputSet.observationG.write2cytoscape(cmfa.indicatorVec, cmfa.mIndicatorQ, inputSet.aSortedProteins)

    return 0

def get_args():
    parser = argparse.ArgumentParser(description='MFA')
    parser.add_argument('-f', '--file', metavar='file',
                        default='', help='CSV input file of protein purifications')
    parser.add_argument('-bp', '--bioplex', metavar='bioplex',
                        default='', help='Indicate if the input is in Bioplex format')
    parser.add_argument('-k', '--max', metavar='numclusters',
                        default='100', help='A maximum number of possible clusters')
    parser.add_argument('-psi', '--ratio', metavar='psi',
                        default='3.4', help='A ratio of log(1-fn)/log(1-fp)')
    return parser.parse_args()

def main():
    args = get_args()
    if (args.file == '' and args.bioplex == ''):
        print('Input file cannot be empty. Require a CSV file of protein purifications.')
        exit()
    
    nK = int(args.max)
    psi = float(args.ratio)

    if args.bioplex != '':
        print('Hello, ' + args.bioplex)
        inputSet = inputBioplex.CInputBioplex(args.bioplex, cpmBioplex.CountBioplexSpoke)
    else:
        print('Hello, ' + args.file)
        inputSet = inputFile.CInputSet(args.file, cpm.CountSpokeModel)
    nLogLikelihood = clustering(inputSet, nK, psi)

if __name__ == '__main__':
    main()



import numpy as np

#
# TODO: cpmFunc can be countSpokeModel or countMatrixModel
#
class CInputSet:

    def __init__(self, filename, cpmFunc):
        super().__init__()

        listBaits = list()
        with open(filename) as fh:
            setProteins = set()
            for line in fh:
                lst = line.rstrip().split(',')
                bait = lst[0]
                listBaits.append(bait)
                setProteins = setProteins.union(set(lst))
            print('Number of proteins ' + str(len(setProteins)))
            fh.close()

        self.aSortedProteins = np.sort(np.array(list(setProteins), dtype='U21'))
        bait_inds = np.searchsorted(self.aSortedProteins, np.array(listBaits, dtype='U21'))
        
        print('Number of purifications ' + str(len(bait_inds)))

        nProteins = len(self.aSortedProteins)
        incidence = np.zeros(shape=(len(bait_inds), nProteins), dtype=int)
        with open(filename) as fh:
            lineCount = 0
            for line in fh:
                lst = line.rstrip().split(',')
                prey_inds = np.searchsorted(self.aSortedProteins, np.array(lst, dtype='U21'))           
                for id in prey_inds:
                    incidence[lineCount][id] = 1
                lineCount += 1
            fh.close()
            
        self.observationG = cpmFunc(nProteins, bait_inds, incidence)

    def writeCluster2File(self, matQ, indVec):
        nRows, nCols = matQ.shape
        with open("out.tab", "w") as fh:
            for i in range(nRows):
                ind = indVec[i]
                fh.write(self.aSortedProteins[i] + '\t' + str(indVec[i]) + '\t' + str(max(matQ[ind])) + '\n')
            fh.close()
        with open("out.csv", "w") as fh:
            for k in range(nCols):
                inds = [i for i in range(nRows) if k == indVec[i]]
                for i in inds:
                    prot = self.aSortedProteins[i].split("__")[0]
                    fh.write(prot + '\t')
                fh.write('\n')    
            fh.close()


import numpy as np
from numba import jit, njit, prange
from numba import float32
from time import time as timer

@njit(parallel=True)
def matmul(A, B, C):
    nCols = len(C)
    for j in prange(nCols):
        tmp = B[:,j]
        C[j] = np.dot(A, tmp)

def test_matmul(A,B):
    nRows = A.shape[0]
    nK = B.shape[1]
    C = np.zeros(nK, dtype=float)
    ts = timer()
    matmul(A, B, C)
    te = timer()
    print("matmul GPU time:", te - ts)

    ts = timer()
    ans = np.matmul(A, B)
    te = timer()
    print("NumPy time:", te - ts)
    np.testing.assert_allclose(ans, C, rtol=1e-5)
    return ans


import numpy as np
from numpy.random import default_rng

import matplotlib.pyplot as plt
import sys
import scipy.special
import scipy.stats

from sklearn import datasets, linear_model
from sklearn.metrics import mean_squared_error, r2_score
from sklearn.linear_model import PoissonRegressor, Ridge
from sklearn.pipeline import Pipeline

from scipy import stats

class CMeanFieldAnnealing:

    def __init__(self, Nproteins, Nk):
        self.lstExpectedLikelihood = []
        self.mIndicatorQ = np.zeros((Nproteins, Nk), dtype=float)
        
    def estimate(self, mObservationG, Nproteins, Nk, psi):

        rng = default_rng()

        # psi = (-np.log(fp) + np.log(1 - fn))/(-np.log(fn) + np.log(1 - fp))
        print('psi = ', psi)

        for i in range(Nproteins):
            self.mIndicatorQ[i,:] = rng.random(Nk)
            self.mIndicatorQ[i,:] /= sum(self.mIndicatorQ[i,:])

        nTemperature = 1000.0
        while nTemperature > 0.0:
            tmp = 1e-2
            mExpectation = 0.0
            nIteration = 0
            while not np.allclose(tmp, mExpectation, 1e-5):
                tmp = mExpectation
                mExpectation = 0.0
                for i in range(Nproteins):
                    # i = np.random.randint(0, Nproteins) # Choose a node at random
                    """ 
                    mLogLikelihood = np.zeros(Nk, dtype=float) # Negative log-likelihood
                    for k in range(Nk):
                        for j in mObservationG.lstAdjacency[i]:
                            t = mObservationG.mTrials[i][j]
                            s = mObservationG.mObserved[i][j]
                            assert(s <= t)
                            mLogLikelihood[k] += (self.mIndicatorQ[j][k]*float(t-s) + (1.0 - self.mIndicatorQ[j][k])*float(s)*psi)
                            ## mLogLikelihood[k] += self.mIndicatorQ[j][k]*(t-s-s*psi)
                    """

                    fn_out = np.dot(mObservationG.mTrials[i] - mObservationG.mObserved[i], self.mIndicatorQ) 
                    fp_out = np.dot(psi*mObservationG.mObserved[i], np.ones((Nproteins, Nk)) - self.mIndicatorQ)

                    mLogLikelihood = fn_out + fp_out

                    # Overflow problem. Need to compute with softmax
                    gamma = nTemperature
                    self.mIndicatorQ[i,:] = scipy.special.softmax(-gamma*mLogLikelihood)
                    ## self.mIndicatorQ[i,:] /= sum(self.mIndicatorQ[i,:])
                nIteration += 1
                mExpectation = np.sum(np.sum(self.mIndicatorQ, axis=0))
            print('Num. Iterations = ', nIteration)    
            print('Temperature: ', nTemperature, 'Expectation = ', mExpectation)
            nTemperature = nTemperature - 100.0
        return self.lstExpectedLikelihood

    ##
    ## Adapt from https://github.com/zib-cmd/cmdtools/blob/dev/src/cmdtools/analysis/optimization.py
    ##
    def inner_simplex_algorithm(self, X):
        """
        Return the transformation A mapping those rows of X
        which span the largest simplex onto the unit simplex.
        """
        ind = self.indexsearch(X)
        return np.linalg.inv(X[ind, :])

    def indexsearch(self, X):
        """ Return the indices to the rows spanning the largest simplex """

        k = np.size(X, axis=1)
        X = X.copy()

        ind = np.zeros(k, dtype=int)
        for j in range(0, k):
            # find the largest row
            rownorm = np.linalg.norm(X, axis=1)
            ind[j] = np.argmax(rownorm)

            if j == 0:
                # translate to origin
                X -= X[ind[j], :]
            else:
                # remove subspace of this row
                X /= rownorm[ind[j]]
                v  = X[ind[j], :]
                X -= np.outer(X.dot(v), v)

        return ind

    def find_lin_dependent(self):
        N = np.size(self.mIndicatorQ, axis=0)
        k = np.size(self.mIndicatorQ, axis=1)
        self.indicatorVec = np.zeros(N, dtype=int)
        ind = self.indexsearch(self.mIndicatorQ)
        for id in ind:
            for i in range(0,N):
                inner_product = np.inner(
                    self.mIndicatorQ[i],
                    self.mIndicatorQ[id])

                norm_i = np.linalg.norm(self.mIndicatorQ[i])
                norm_j = np.linalg.norm(self.mIndicatorQ[id])

                distance = np.abs(inner_product - norm_i*norm_j)
                if distance < 1E-5:
                    self.indicatorVec[i] = id
            self.indicatorVec[id] = id

    def find_argmax(self):
        N = np.size(self.mIndicatorQ, axis=0)
        k = np.size(self.mIndicatorQ, axis=1)
        self.indicatorVec = np.argmax(self.mIndicatorQ, axis=1)

    def computeErrorRate(self, mObservationG, Nproteins):
        
        # self.find_lin_dependent()
        self.find_argmax()

        rnk = np.linalg.matrix_rank(self.mIndicatorQ)
        print("Indicator matrix had rank = " + str(rnk))
        nClusters = len(np.unique(self.indicatorVec))
        print("Number of clusters used: " + str(nClusters))

        countFn = 0
        countFp = 0
        sumSameCluster = 0
        sumDiffCluster = 0
        for i in range(Nproteins):
            for j in mObservationG.lstAdjacency[i]:
                t = mObservationG.mTrials[i][j]
                s = mObservationG.mObserved[i][j]
                assert(s <= t)
                if (self.indicatorVec[i] == self.indicatorVec[j]):
                    countFn += (t - s)
                    sumSameCluster += t
                else:
                    countFp += s
                    sumDiffCluster += t

        fn = 0.0
        fp = 0.0
        if (sumSameCluster > 0):
            fn = float(countFn)/float(sumSameCluster)
        if (sumDiffCluster > 0):
            fp = float(countFp)/float(sumDiffCluster)
        return (fn, fp)

    def estimator_summary(self, regr, y_actual, y_pred):
        # The coefficients
        print("Coefficients: \n", regr.coef_)
        # The mean squared error
        print("Mean squared error: %.2f" % mean_squared_error(y_actual, y_pred))
        # The coefficient of determination: 1 is perfect prediction
        print("Coefficient of determination: %.2f" % r2_score(y_actual, y_pred))
        
    def computeResidues(self, mObservationG, Nproteins, Nk):
    
        (fn, fp) = self.computeErrorRate(mObservationG, Nproteins)

        # Filter singletons
        clusters = []
        singletons = []
        for k in range(Nk): 
            if (sum(self.indicatorVec == k) > 1):
                clusters.append(k)
            else:
                singletons.append(k)
        setProteins = set()
        indicators = dict()
        for i in range(Nproteins):
            if self.indicatorVec[i] in clusters: 
                setProteins.add(i)
                indicators[i] = self.indicatorVec[i]

        countFn = np.zeros(Nk)
        countFp = np.zeros(Nk)
        trialFn = np.zeros(Nk)
        trialFp = np.zeros(Nk)
        for i in setProteins:
            cl = indicators[i]    
            for j in mObservationG.lstAdjacency[i]:
                t = mObservationG.mTrials[i][j]
                s = mObservationG.mObserved[i][j]
                if (self.indicatorVec[i] == cl) and (self.indicatorVec[j] == cl):      
                    trialFn[cl] += t
                    countFn[cl] += (t - s)
                if (self.indicatorVec[i] == cl) and (self.indicatorVec[j] != cl):
                    countFp[cl] += s
                    trialFp[cl] += t

        self.expectedErrors = np.floor(fn*trialFn) + np.floor(fp*trialFp)
        self.mResidues = countFn + countFp

        expectedErrors = list( self.expectedErrors[i] for i in clusters)
        residues = list( self.mResidues[i] for i in clusters)
        result = scipy.stats.ks_2samp(residues, expectedErrors)
        print(result)


        self.expectedErrors = expectedErrors
        self.mResidues = residues

        """
        # Ordinary Least Square
        est = sm.OLS(residues,  X)
        est2 = est.fit()
        print(est2.summary())

        glm_poisson = sm.GLM(residues, X, family=sm.families.Poisson(sm.families.links.log()))
        glm_results = glm_poisson.fit()
        print(glm_results.summary())
        """

        X = np.reshape(expectedErrors, (-1,1))
        # Create linear regression object
        regr = linear_model.LinearRegression()
        # Train the model using the training sets
        regr.fit(X, residues)
        # Make predictions using the testing set
        y_pred = regr.predict(X)
        print("Linear Regression evaluation:")
        self.estimator_summary(regr, residues, y_pred)
        
        return (fn, fp)

    def computeEntropy(self, Nproteins, Nk):
        self.mEntropy = np.zeros(Nproteins, dtype=float)
        for i in range(Nproteins):
            for k in range(Nk):
                p = self.mIndicatorQ[i][k]
                if (p > 0):
                    self.mEntropy[i] += (-p*np.log(p))
        
    #
    # https://www.researchgate.net/publication/343831904_NumPy_SciPy_Recipes_for_Data_Science_Projections_onto_the_Standard_Simplex
    #
    def indicator2simplex(self):
        m,n = self.mIndicatorQ.shape
        matS = np.sort(self.mIndicatorQ, axis=0)
        matC = np.cumsum(matS, axis=0) - 1.
        matH = matS - matC / (np.arange(m) + 1).reshape(m,1)
        matH[matH<=0] = np.inf

        r = np.argmin(matH, axis=0)
        t = matC[r,np.arange(n)] / (r + 1)

        matY = self.mIndicatorQ - t
        matY[matY < 0] = 0
        return matY

    def clusterImage(self, matQ):
        m,n = matQ.shape
        vecArgMax = np.argmax(matQ,axis=1)
        vecIndices = np.argsort(vecArgMax)
        matImage = np.zeros((m,n), dtype=int)
        for i in range(len(vecIndices)):
            k = vecArgMax[vecIndices[i]]
            matImage[i][k] = 255
        return matImage



import pandas as pd
import numpy as np

class CountBioplexSpoke:

    def __init__(self, filePath):

        df = pd.read_csv(filePath, sep='\t')

        df_filtered = df[df.apply(lambda x: not x['bait_symbol'].isnumeric() and x['bait_symbol'] != "nan", axis=1)]
        df_filtered = df[df.apply(lambda x: isinstance(x['symbol'], str) and not x['symbol'].isnumeric(), axis=1)]

        bait_list = np.array(df_filtered['bait_symbol'], dtype='U21')
        prey_list = np.array(df_filtered['symbol'], dtype='U21')

        proteins_list = np.append(bait_list, prey_list)
            
        self.nProteins = len(np.unique(proteins_list))
        nProteins = self.nProteins
        
        npSortedProteins = np.sort(np.unique(proteins_list))  # sorted proteins list
        bait_inds = np.searchsorted(npSortedProteins,bait_list) # array([[2, 1, 0, 2,...]])
        prey_inds = np.searchsorted(npSortedProteins,prey_list)
        # proteins_list = list(proteins_set)
        self.mObserved = np.zeros(shape=(nProteins, nProteins), dtype=int)
        nrows, ncols = df.shape
        for i,j in zip(bait_inds, prey_inds):
            # bait = df.loc[row,'bait_symbol']
            # prey = df.loc[row,'symbol']
            self.mObserved[i][j] += 1
            self.mObserved[j][i] += 1

        self.mTrials = np.zeros(shape=(nProteins, nProteins), dtype=int)
        bincounts = np.bincount(bait_inds)
        for i, k in enumerate(bincounts):
            self.mTrials[i,:] += k*np.ones(nProteins, dtype=np.int32)
            self.mTrials[:,i] += k*np.ones(nProteins, dtype=np.int32)
            
        for i in range(nProteins):
            assert(np.sum(self.mTrials[i,:]) == np.sum(self.mTrials[:,i]))

        #
        # Create the adjacency list
        #
        self.lstAdjacency = {}
        for i in np.arange(nProteins):
            self.lstAdjacency[i] = []
            for j in np.arange(nProteins):
                t = self.mTrials[i][j]
                if (i < j):
                    s = self.mObserved[i][j] 
                else:
                    s = self.mObserved[j][i] 
                if (i != j and t > 0):
                    assert(s <= t)
                    self.lstAdjacency[i].append(j)

    def write2cytoscape(self, indicators, matQ, vecProteins):
        nRows, nCols = matQ.shape
        with open("out.sif", "w") as fh:
            for i in np.arange(nRows):
                for j in self.lstAdjacency[i]:
                    t = self.mTrials[i][j]
                    if (i >= j):
                        continue
                    if (t > 0 and indicators[i] == indicators[j]):
                        fh.write(str(vecProteins[i]) + '\t' + str(indicators[i]) + '\t' + str(vecProteins[j]) + '\n')
            fh.close()


import numpy as np
import pandas as pd

#
# TODO: cpmFunc can be countSpokeModel or countMatrixModel
#
class CInputBioplex:

    def __init__(self, filePath, cpmFunc):
        super().__init__()

        df = pd.read_csv(filePath, sep='\t')

        df_filtered = df[df.apply(lambda x: not x['bait_symbol'].isnumeric() and x['bait_symbol'] != "nan", axis=1)]
        df_filtered = df[df.apply(lambda x: isinstance(x['symbol'], str) and not x['symbol'].isnumeric(), axis=1)]

        bait_list = np.array(df_filtered['bait_symbol'], dtype='U21')
        prey_list = np.array(df_filtered['symbol'], dtype='U21')

        proteins_list = np.append(bait_list, prey_list)
            
        self.nProteins = len(np.unique(proteins_list))
        nProteins = self.nProteins

        print('Number of baits = ', len(np.unique(bait_list)))
        print('Number of preys = ', len(np.unique(prey_list)))
        print('Number of proteins = ', nProteins)

        self.aSortedProteins = np.sort(np.unique(proteins_list))  # sorted proteins list
        
        for v in self.aSortedProteins[:100]:
            print(v)

        self.observationG = cpmFunc(filePath)

    def writeCluster2File(self, matQ, indVec):
        nRows, nCols = matQ.shape
        with open("bioplex_out.tab", "w") as fh:
            for i in range(nRows):
                ind = indVec[i]
                fh.write(str(self.aSortedProteins[i]) + '\t' + str(indVec[i]) + '\t' + str(max(matQ[ind])) + '\n')
            fh.close()
        with open("bioplex_out.csv", "w") as fh:
            for k in range(nCols):
                inds = list(i for i in range(nRows) if k == indVec[i])
                for i in inds:
                    fh.write(self.aSortedProteins[i] + '\t')
                fh.write('\n')
            fh.close()


from setuptools import find_packages, setup

setup(name='braid-mrf',
      version='1.0.9',
      description='Predicting protein complexes',
      author='Wasinee Rungsarityotin',
      author_email='wasinees@gmail.com',
      url='https://github.com/wasineer-dev/braid.git',
      packages=['braidmrf', 'braidmrf.meanfield', 'braidmrf.inputFile', 'braidmrf.observation'],
     )